"""Portions of this code are derived from the Apache 2.0 licensed project mlflow (https://mlflow.org/).,
with modifications made by Scaleout Systems AB.
Copyright (c) 2018 Databricks, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import yaml

from fedn.utils import PYTHON_VERSION

_REQUIREMENTS_FILE_NAME = "requirements.txt"
_PYTHON_ENV_FILE_NAME = "python_env.yaml"


class _PythonEnv:
    BUILD_PACKAGES = ("pip", "setuptools", "wheel")

    def __init__(self, name=None, python=None, build_dependencies=None, dependencies=None):
        """Represents environment information for FEDn compute packages.

        Args:
        ----
            name: Name of environment. If unspecified, defaults to fedn_env
            python: Python version for the environment. If unspecified, defaults to the current
                Python version.
            build_dependencies: List of build dependencies for the environment that must
                be installed before installing ``dependencies``. If unspecified,
                defaults to an empty list.
            dependencies: List of dependencies for the environment. If unspecified, defaults to
                an empty list.

        """
        if name is not None and not isinstance(name, str):
            raise TypeError(f"`name` must be a string but got {type(name)}")
        if python is not None and not isinstance(python, str):
            raise TypeError(f"`python` must be a string but got {type(python)}")
        if build_dependencies is not None and not isinstance(build_dependencies, list):
            raise TypeError(f"`build_dependencies` must be a list but got {type(build_dependencies)}")
        if dependencies is not None and not isinstance(dependencies, list):
            raise TypeError(f"`dependencies` must be a list but got {type(dependencies)}")
        self.name = name or "fedn_env"
        self.python = python or PYTHON_VERSION
        self.build_dependencies = build_dependencies or []
        self.dependencies = dependencies or []

    def __str__(self):
        return str(self.to_dict())

    @classmethod
    def current(cls):
        return cls(
            python=PYTHON_VERSION,
            build_dependencies=cls.get_current_build_dependencies(),
            dependencies=[f"-r {_REQUIREMENTS_FILE_NAME}"],
        )

    @staticmethod
    def _get_package_version(package_name):
        try:
            return __import__(package_name).__version__
        except (ImportError, AttributeError, AssertionError):
            return None

    @staticmethod
    def get_current_build_dependencies():
        build_dependencies = []
        for package in _PythonEnv.BUILD_PACKAGES:
            version = _PythonEnv._get_package_version(package)
            dep = (package + "==" + version) if version else package
            build_dependencies.append(dep)
        return build_dependencies

    def to_dict(self):
        return self.__dict__.copy()

    @classmethod
    def from_dict(cls, dct):
        return cls(**dct)

    def to_yaml(self, path):
        with open(path, "w") as f:
            # Exclude None and empty lists
            data = {k: v for k, v in self.to_dict().items() if v}
            yaml.safe_dump(data, f, sort_keys=False, default_flow_style=False)

    @classmethod
    def from_yaml(cls, path):
        with open(path) as f:
            return cls.from_dict(yaml.safe_load(f))

    @staticmethod
    def get_dependencies_from_conda_yaml(path):
        raise NotImplementedError

    @classmethod
    def from_conda_yaml(cls, path):
        return cls.from_dict(cls.get_dependencies_from_conda_yaml(path))



# This file contains the Connector class for assigning client to the FEDn network via the discovery service (REST-API).
# The Connector class is used by the Client class in fedn/network/clients/client.py.
# Once assigned, the client will retrieve combiner assignment from the discovery service.
# The discovery service will also add the client to the statestore.
#
#
import enum

import requests

from fedn.common.config import FEDN_AUTH_REFRESH_TOKEN, FEDN_AUTH_REFRESH_TOKEN_URI, FEDN_AUTH_SCHEME, FEDN_CUSTOM_URL_PREFIX
from fedn.common.log_config import logger


class Status(enum.Enum):
    """Enum for representing the status of a client assignment."""

    Unassigned = 0
    Assigned = 1
    TryAgain = 2
    UnAuthorized = 3
    UnMatchedConfig = 4


class ConnectorClient:
    """Connector for assigning client to a combiner in the FEDn network.

    :param host: host of discovery service
    :type host: str
    :param port: port of discovery service
    :type port: int
    :param token: token for authentication
    :type token: str
    :param name: name of client
    :type name: str
    :param remote_package: True if remote package is used, False if local
    :type remote_package: bool
    :param force_ssl: True if https is used, False if http
    :type force_ssl: bool
    :param verify: True if certificate is verified, False if not
    :type verify: bool
    :param combiner: name of preferred combiner
    :type combiner: str
    :param id: id of client
    """

    def __init__(self, host, port, token, name, remote_package, force_ssl=False, verify=False, combiner=None, id=None):
        self.host = host
        self.port = port
        self.token = token
        self.name = name
        self.verify = verify
        self.preferred_combiner = combiner
        self.id = id
        self.package = "remote" if remote_package else "local"

        # for https we assume a an ingress handles permanent redirect (308)
        if force_ssl:
            self.prefix = "https://"
        else:
            self.prefix = "http://"
        if self.port:
            self.connect_string = "{}{}:{}".format(self.prefix, self.host, self.port)
        else:
            self.connect_string = "{}{}".format(self.prefix, self.host)

        logger.info("Setting connection string to {}.".format(self.connect_string))

    def assign(self):
        """Connect client to FEDn network discovery service, ask for combiner assignment.

        :return: Tuple with assingment status, combiner connection information if sucessful, else None.
        :rtype: tuple(:class:`fedn.network.clients.connect.Status`, str)
        """
        try:
            retval = None
            payload = {"name": self.name, "client_id": self.id, "preferred_combiner": self.preferred_combiner}
            retval = requests.post(
                self.connect_string + FEDN_CUSTOM_URL_PREFIX + "/add_client",
                json=payload,
                verify=self.verify,
                allow_redirects=True,
                headers={"Authorization": f"{FEDN_AUTH_SCHEME} {self.token}"},
            )
        except Exception as e:
            logger.debug("***** {}".format(e))
            return Status.Unassigned, {}

        if retval.status_code == 400:
            # Get error messange from response
            reason = retval.json()["message"]
            return Status.UnMatchedConfig, reason

        if retval.status_code == 401:
            if "message" in retval.json():
                reason = retval.json()["message"]
                logger.warning(reason)
                if reason == "Token expired":
                    status_code = self.refresh_token()
                    if status_code >= 200 and status_code < 204:
                        logger.info("Token refreshed.")
                        return Status.TryAgain, reason
                    else:
                        return Status.UnAuthorized, "Could not refresh token"
            reason = "Unauthorized connection to reducer, make sure the correct token is set"
            return Status.UnAuthorized, reason

        if retval.status_code >= 200 and retval.status_code < 204:
            if retval.json()["status"] == "retry":
                if "message" in retval.json():
                    reason = retval.json()["message"]
                else:
                    reason = "Reducer was not ready. Try again later."

                return Status.TryAgain, reason

            reducer_package = retval.json()["package"]
            if reducer_package != self.package:
                reason = "Unmatched config of compute package between client and reducer.\n" + "Reducer uses {} package and client uses {}.".format(
                    reducer_package, self.package
                )
                return Status.UnMatchedConfig, reason

            return Status.Assigned, retval.json()

        return Status.Unassigned, None

    def refresh_token(self):
        """Refresh client token.

        :return: Tuple with assingment status, combiner connection information if sucessful, else None.
        :rtype: tuple(:class:`fedn.network.clients.connect.Status`, str)
        """
        if not FEDN_AUTH_REFRESH_TOKEN_URI or not FEDN_AUTH_REFRESH_TOKEN:
            logger.error("No refresh token URI/Token set, cannot refresh token.")
            return 401

        payload = requests.post(FEDN_AUTH_REFRESH_TOKEN_URI, verify=self.verify, allow_redirects=True, json={"refresh": FEDN_AUTH_REFRESH_TOKEN})
        self.token = payload.json()["access"]
        return payload.status_code



from fedn.utils.dist import get_version
import click

CONTEXT_SETTINGS = dict(
    # Support -h as a shortcut for --help
    help_option_names=["-h", "--help"],
)

version=get_version("fedn")


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(version)
@click.pass_context
def main(ctx):
    """:param ctx:"""
    ctx.obj = dict()



import os
from io import BytesIO

import fedn.network.grpc.fedn_pb2 as fedn
from fedn.common.log_config import logger
from fedn.network.storage.models.modelstorage import ModelStorage

CHUNK_SIZE = 1024 * 1024


class TempModelStorage(ModelStorage):
    """Class for managing local temporary models on file on combiners."""

    def __init__(self):
        self.default_dir = os.environ.get("FEDN_MODEL_DIR", "/tmp/models")  # set default to tmp
        if not os.path.exists(self.default_dir):
            os.makedirs(self.default_dir)

        self.models = {}
        self.models_metadata = {}

    def exist(self, model_id):
        if model_id in self.models.keys():
            return True
        return False

    def get(self, model_id):
        try:
            if self.models_metadata[model_id] != fedn.ModelStatus.OK:
                logger.warning("File not ready! Try again")
                return None
        except KeyError:
            logger.error("No such model has been made available yet!")
            return None

        obj = BytesIO()
        with open(os.path.join(self.default_dir, str(model_id)), "rb") as f:
            obj.write(f.read())

        obj.seek(0, 0)
        return obj

    def get_ptr(self, model_id):
        """:param model_id:
        :return:
        """
        try:
            f = self.models[model_id]["file"]
        except KeyError:
            f = open(os.path.join(self.default_dir, str(model_id)), "wb")

        self.models[model_id] = {"file": f}
        return self.models[model_id]["file"]

    def get_model_metadata(self, model_id):
        try:
            status = self.models_metadata[model_id]
        except KeyError:
            status = fedn.ModelStatus.UNKNOWN
        return status

    def set_model_metadata(self, model_id, model_metadata):
        self.models_metadata.update({model_id: model_metadata})

    # Delete model from disk
    def delete(self, model_id):
        try:
            os.remove(os.path.join(self.default_dir, str(model_id)))
            logger.info("TEMPMODELSTORAGE: Deleted model with id: {}".format(model_id))
            # Delete id from metadata and models dict
            del self.models_metadata[model_id]
            del self.models[model_id]
        except FileNotFoundError:
            logger.error("Could not delete model from disk. File not found!")
            return False
        return True

    # Delete all models from disk
    def delete_all(self):
        ids_pop = []
        for model_id in self.models.keys():
            try:
                os.remove(os.path.join(self.default_dir, str(model_id)))
                logger.info("TEMPMODELSTORAGE: Deleted model with id: {}".format(model_id))
                # Add id to list of ids to pop/delete from metadata and models dict
                ids_pop.append(model_id)
            except FileNotFoundError:
                logger.error("TEMPMODELSTORAGE: Could not delete model {} from disk. File not found!".format(model_id))
        # Remove id from metadata and models dict
        for model_id in ids_pop:
            del self.models_metadata[model_id]
            del self.models[model_id]
        return True



import os
import uuid
from abc import ABC, abstractmethod
from time import sleep

import fedn.utils.helpers.helpers
from fedn.common.log_config import logger
from fedn.network.api.network import Network
from fedn.network.combiner.interfaces import CombinerUnavailableError
from fedn.network.combiner.roundhandler import RoundConfig
from fedn.network.state import ReducerState
from fedn.network.storage.s3.repository import Repository

# Maximum number of tries to connect to statestore and retrieve storage configuration
MAX_TRIES_BACKEND = os.getenv("MAX_TRIES_BACKEND", 10)


class UnsupportedStorageBackend(Exception):
    pass


class MisconfiguredStorageBackend(Exception):
    pass


class MisconfiguredHelper(Exception):
    pass


class ControlBase(ABC):
    """Base class and interface for a global controller.
        Override this class to implement a global training strategy (control).

    :param statestore: The statestore object.
    :type statestore: :class:`fedn.network.statestore.statestorebase.StateStoreBase`
    """

    @abstractmethod
    def __init__(self, statestore):
        """Constructor."""
        self._state = ReducerState.setup

        self.statestore = statestore
        if self.statestore.is_inited():
            self.network = Network(self, statestore)

        try:
            not_ready = True
            tries = 0
            while not_ready:
                storage_config = self.statestore.get_storage_backend()
                if storage_config:
                    not_ready = False
                else:
                    logger.warning("Storage backend not configured, waiting...")
                    sleep(5)
                    tries += 1
                    if tries > MAX_TRIES_BACKEND:
                        raise Exception
        except Exception:
            logger.error("Failed to retrive storage configuration, exiting.")
            raise MisconfiguredStorageBackend()

        if storage_config["storage_type"] == "S3":
            self.model_repository = Repository(storage_config["storage_config"])
        else:
            logger.error("Unsupported storage backend, exiting.")
            raise UnsupportedStorageBackend()

        if self.statestore.is_inited():
            self._state = ReducerState.idle

    @abstractmethod
    def session(self, config):
        pass

    @abstractmethod
    def round(self, config, round_number):
        pass

    @abstractmethod
    def reduce(self, combiners):
        pass

    def get_helper(self):
        """Get a helper instance from global config.

        :return: Helper instance.
        :rtype: :class:`fedn.utils.plugins.helperbase.HelperBase`
        """
        helper_type = self.statestore.get_helper()
        helper = fedn.utils.helpers.helpers.get_helper(helper_type)
        if not helper:
            raise MisconfiguredHelper("Unsupported helper type {}, please configure compute_package.helper !".format(helper_type))
        return helper

    def get_state(self):
        """Get the current state of the controller.

        :return: The current state.
        :rtype: :class:`fedn.network.state.ReducerState`
        """
        return self._state

    def idle(self):
        """Check if the controller is idle.

        :return: True if idle, False otherwise.
        :rtype: bool
        """
        if self._state == ReducerState.idle:
            return True
        else:
            return False

    def get_model_info(self):
        """:return:"""
        return self.statestore.get_model_trail()

    # TODO: remove use statestore.get_events() instead
    def get_events(self):
        """:return:"""
        return self.statestore.get_events()

    def get_latest_round_id(self):
        last_round = self.statestore.get_latest_round()
        if not last_round:
            return 0
        else:
            return last_round["round_id"]

    def get_latest_round(self):
        round = self.statestore.get_latest_round()
        return round

    def get_compute_package_name(self):
        """:return:"""
        definition = self.statestore.get_compute_package()
        if definition:
            try:
                package_name = definition["storage_file_name"]
                return package_name
            except (IndexError, KeyError):
                logger.error("No context filename set for compute context definition")
                return None
        else:
            return None

    def set_compute_package(self, filename, path):
        """Persist the configuration for the compute package."""
        self.model_repository.set_compute_package(filename, path)

    def get_compute_package(self, compute_package=""):
        """:param compute_package:
        :return:
        """
        if compute_package == "":
            compute_package = self.get_compute_package_name()
        if compute_package:
            return self.model_repository.get_compute_package(compute_package)
        else:
            return None

    def create_session(self, config: RoundConfig, status: str = "Initialized") -> None:
        """Initialize a new session in backend db."""
        if "session_id" not in config.keys():
            session_id = uuid.uuid4()
            config["session_id"] = str(session_id)
        else:
            session_id = config["session_id"]

        self.statestore.create_session(id=session_id)
        self.statestore.set_session_config(session_id, config)
        self.statestore.set_session_status(session_id, status)

    def set_session_status(self, session_id, status):
        """Set the round round stats.

        :param round_id: The round unique identifier
        :type round_id: str
        :param status: The status
        :type status: str
        """
        self.statestore.set_session_status(session_id, status)

    def get_session_status(self, session_id):
        """Get the status of a session.

        :param session_id: The session unique identifier
        :type session_id: str
        :return: The status
        :rtype: str
        """
        return self.statestore.get_session_status(session_id)

    def set_session_config(self, session_id: str, config: dict):
        """Set the model id for a session.

        :param session_id: The session unique identifier
        :type session_id: str
        :param config: The session config
        :type config: dict
        """
        self.statestore.set_session_config_v2(session_id, config)

    def create_round(self, round_data):
        """Initialize a new round in backend db."""
        self.statestore.create_round(round_data)

    def set_round_data(self, round_id, round_data):
        """Set round data.

        :param round_id: The round unique identifier
        :type round_id: str
        :param round_data: The status
        :type status: dict
        """
        self.statestore.set_round_data(round_id, round_data)

    def set_round_status(self, round_id, status):
        """Set the round round stats.

        :param round_id: The round unique identifier
        :type round_id: str
        :param status: The status
        :type status: str
        """
        self.statestore.set_round_status(round_id, status)

    def set_round_config(self, round_id, round_config: RoundConfig):
        """Upate round in backend db.

        :param round_id: The round unique identifier
        :type round_id: str
        :param round_config: The round configuration
        :type round_config: dict
        """
        self.statestore.set_round_config(round_id, round_config)

    def request_model_updates(self, combiners):
        """Ask Combiner server to produce a model update.

        :param combiners: A list of combiners
        :type combiners: tuple (combiner, combiner_round_config)
        """
        cl = []
        for combiner, combiner_round_config in combiners:
            response = combiner.submit(combiner_round_config)
            cl.append((combiner, response))
        return cl

    def commit(self, model_id, model=None, session_id=None):
        """Commit a model to the global model trail. The model commited becomes the lastest consensus model.

        :param model_id: Unique identifier for the model to commit.
        :type model_id: str (uuid)
        :param model: The model object to commit
        :type model: BytesIO
        :param session_id: Unique identifier for the session
        :type session_id: str
        """
        helper = self.get_helper()
        if model is not None:
            logger.info("Saving model file temporarily to disk...")
            outfile_name = helper.save(model)
            logger.info("CONTROL: Uploading model to Minio...")
            model_id = self.model_repository.set_model(outfile_name, is_file=True)

            logger.info("CONTROL: Deleting temporary model file...")
            os.unlink(outfile_name)

        logger.info("Committing model {} to global model trail in statestore...".format(model_id))
        self.statestore.set_latest_model(model_id, session_id)

    def get_combiner(self, name):
        for combiner in self.network.get_combiners():
            if combiner.name == name:
                return combiner
        return None

    def get_participating_combiners(self, combiner_round_config):
        """Assemble a list of combiners able to participate in a round as
        descibed by combiner_round_config.
        """
        combiners = []
        for combiner in self.network.get_combiners():
            try:
                # Current gRPC endpoint only returns active clients (both trainers and validators)
                nr_active_clients = len(combiner.list_active_clients())
            except CombinerUnavailableError:
                self._handle_unavailable_combiner(combiner)
                continue

            clients_required = int(combiner_round_config["clients_required"])
            is_participating = self.evaluate_round_participation_policy(clients_required, nr_active_clients)
            if is_participating:
                combiners.append((combiner, combiner_round_config))
        return combiners

    def evaluate_round_participation_policy(self, clients_required: int, nr_active_clients: int) -> bool:
        """Evaluate policy for combiner round-participation.
        A combiner participates if it is responsive and reports enough
        active clients to participate in the round.
        """
        if clients_required <= nr_active_clients:
            return True
        else:
            return False

    def evaluate_round_start_policy(self, combiners: list):
        """Check if the policy to start a round is met.

        :param combiners: A list of combiners
        :type combiners: list
        :return: True if the round policy is mer, otherwise False
        :rtype: bool
        """
        if len(combiners) > 0:
            return True
        else:
            return False

    def evaluate_round_validity_policy(self, round):
        """Check if the round is valid.

        At the end of the round, before committing a model to the global model trail,
        we check if the round validity policy has been met. This can involve
        e.g. asserting that a certain number of combiners have reported in an
        updated model, or that criteria on model performance have been met.

        :param round: The round object
        :rtype round: dict
        :return: True if the policy is met, otherwise False
        :rtype: bool
        """
        model_ids = []
        for combiner in round["combiners"]:
            try:
                model_ids.append(combiner["model_id"])
            except KeyError:
                pass

        if len(model_ids) == 0:
            return False

        return True

    def state(self):
        """Get the current state of the controller.

        :return: The state
        :rype: str
        """
        return self._state



from typing import Any, Dict, Generic, List, Tuple, TypeVar

import pymongo
from bson import ObjectId
from pymongo.database import Database

from .shared import EntityNotFound, from_document

T = TypeVar("T")


class Store(Generic[T]):
    def __init__(self, database: Database, collection: str):
        self.database = database
        self.collection = collection

    def get(self, id: str, use_typing: bool = False) -> T:
        """Get an entity by id
        param id: The id of the entity
            type: str
        param use_typing: Whether to return the entity as a typed object or as a dict
            type: bool
        return: The entity
        """
        id_obj = ObjectId(id)
        document = self.database[self.collection].find_one({"_id": id_obj})

        if document is None:
            raise EntityNotFound(f"Entity with id {id} not found")

        return from_document(document) if not use_typing else document

    def update(self, id: str, item: T) -> Tuple[bool, Any]:
        try:
            result = self.database[self.collection].update_one({"_id": ObjectId(id)}, {"$set": item})
            if result.modified_count == 1:
                document = self.database[self.collection].find_one({"_id": ObjectId(id)})
                return True, from_document(document)
            else:
                return False, "Entity not found"
        except Exception as e:
            return False, str(e)

    def add(self, item: T) -> Tuple[bool, Any]:
        try:
            result = self.database[self.collection].insert_one(item)
            id = result.inserted_id
            document = self.database[self.collection].find_one({"_id": id})
            return True, from_document(document)
        except Exception as e:
            return False, str(e)

    def delete(self, id: str) -> bool:
        pass

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, use_typing: bool = False, **kwargs) -> Dict[int, List[T]]:
        """List entities
        param limit: The maximum number of entities to return
            type: int
        param skip: The number of entities to skip
            type: int
        param sort_key: The key to sort by
            type: str
        param sort_order: The order to sort by
            type: pymongo.DESCENDING | pymongo.ASCENDING
        param use_typing: Whether to return the entities as typed objects or as dicts
            type: bool
        param kwargs: Additional query parameters
            type: dict
            example: {"key": "models"}
        return: A dictionary with the count and the result
        """
        cursor = self.database[self.collection].find(kwargs).sort(sort_key, sort_order).skip(skip or 0).limit(limit or 0)

        count = self.database[self.collection].count_documents(kwargs)

        result = [document for document in cursor] if use_typing else [from_document(document) for document in cursor]

        return {
            "count": count,
            "result": result
        }

    def count(self, **kwargs) -> int:
        """Count entities
        param kwargs: Additional query parameters
            type: dict
            example: {"key": "models"}
        return: The count (int)
        """
        return self.database[self.collection].count_documents(kwargs)



import uuid

import click
import requests

from fedn.common.exceptions import InvalidClientConfig
from fedn.network.clients.client import Client

from .main import main
from .shared import CONTROLLER_DEFAULTS, apply_config, get_api_url, get_token, print_response


def validate_client_config(config):
    """Validate client configuration.

    :param config: Client config (dict).
    """
    try:
        if config["discover_host"] is None or config["discover_host"] == "":
            raise InvalidClientConfig("Missing required configuration: discover_host")
        if "discover_port" not in config.keys():
            config["discover_port"] = None
    except Exception:
        raise InvalidClientConfig("Could not load config from file. Check config")


@main.group("client")
@click.pass_context
def client_cmd(ctx):
    """:param ctx:
    """
    pass


@click.option("-p", "--protocol", required=False, default=CONTROLLER_DEFAULTS["protocol"], help="Communication protocol of controller (api)")
@click.option("-H", "--host", required=False, default=CONTROLLER_DEFAULTS["host"], help="Hostname of controller (api)")
@click.option("-P", "--port", required=False, default=CONTROLLER_DEFAULTS["port"], help="Port of controller (api)")
@click.option("-t", "--token", required=False, help="Authentication token")
@click.option("--n_max", required=False, help="Number of items to list")
@client_cmd.command("list")
@click.pass_context
def list_clients(ctx, protocol: str, host: str, port: str, token: str = None, n_max: int = None):
    """Return:
    ------
    - count: number of clients
    - result: list of clients

    """
    url = get_api_url(protocol=protocol, host=host, port=port, endpoint="clients")
    headers = {}

    if n_max:
        headers["X-Limit"] = n_max

    _token = get_token(token)

    if _token:
        headers["Authorization"] = _token

    click.echo(f"\nListing clients: {url}\n")
    click.echo(f"Headers: {headers}")

    try:
        response = requests.get(url, headers=headers)
        print_response(response, "clients")
    except requests.exceptions.ConnectionError:
        click.echo(f"Error: Could not connect to {url}")


@client_cmd.command("start")
@click.option("-d", "--discoverhost", required=False, help="Hostname for discovery services(reducer).")
@click.option("-p", "--discoverport", required=False, help="Port for discovery services (reducer).")
@click.option("--token", required=False, help="Set token provided by reducer if enabled")
@click.option("-n", "--name", required=False, default="client" + str(uuid.uuid4())[:8])
@click.option("-i", "--client_id", required=False)
@click.option("--local-package", is_flag=True, help="Enable local compute package")
@click.option("--force-ssl", is_flag=True, help="Force SSL/TLS for REST service")
@click.option("-u", "--dry-run", required=False, default=False)
@click.option("-s", "--secure", required=False, default=False)
@click.option("-pc", "--preshared-cert", required=False, default=False)
@click.option("-v", "--verify", is_flag=True, help="Verify SSL/TLS for REST service")
@click.option("-c", "--preferred-combiner", required=False, default=False)
@click.option("-va", "--validator", required=False, default=True)
@click.option("-tr", "--trainer", required=False, default=True)
@click.option("-in", "--init", required=False, default=None, help="Set to a filename to (re)init client from file state.")
@click.option("-l", "--logfile", required=False, default=None, help="Set logfile for client log to file.")
@click.option("--heartbeat-interval", required=False, default=2)
@click.option("--reconnect-after-missed-heartbeat", required=False, default=30)
@click.option("--verbosity", required=False, default="INFO", type=click.Choice(["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"], case_sensitive=False))
@click.pass_context
def client_cmd(
    ctx,
    discoverhost,
    discoverport,
    token,
    name,
    client_id,
    local_package,
    force_ssl,
    dry_run,
    secure,
    preshared_cert,
    verify,
    preferred_combiner,
    validator,
    trainer,
    init,
    logfile,
    heartbeat_interval,
    reconnect_after_missed_heartbeat,
    verbosity,
):
    """:param ctx:
    :param discoverhost:
    :param discoverport:
    :param token:
    :param name:
    :param client_id:
    :param remote:
    :param dry_run:
    :param secure:
    :param preshared_cert:
    :param verify_cert:
    :param preferred_combiner:
    :param init:
    :param logfile:
    :param hearbeat_interval
    :param reconnect_after_missed_heartbeat
    :param verbosity
    :return:
    """
    remote = False if local_package else True
    config = {
        "discover_host": discoverhost,
        "discover_port": discoverport,
        "token": token,
        "name": name,
        "client_id": client_id,
        "remote_compute_context": remote,
        "force_ssl": force_ssl,
        "dry_run": dry_run,
        "secure": secure,
        "preshared_cert": preshared_cert,
        "verify": verify,
        "preferred_combiner": preferred_combiner,
        "validator": validator,
        "trainer": trainer,
        "logfile": logfile,
        "heartbeat_interval": heartbeat_interval,
        "reconnect_after_missed_heartbeat": reconnect_after_missed_heartbeat,
        "verbosity": verbosity,
    }

    if init:
        apply_config(init, config)
        click.echo(f"\nClient configuration loaded from file: {init}")
        click.echo("Values set in file override defaults and command line arguments...\n")

    try:
        validate_client_config(config)
    except InvalidClientConfig as e:
        click.echo(f"Error: {e}")
        return

    client = Client(config)
    client.run()



import os
import shutil
import uuid

import click
import yaml

from fedn.common.exceptions import InvalidClientConfig
from fedn.common.log_config import logger
from fedn.network.clients.client import Client
from fedn.network.combiner.combiner import Combiner
from fedn.utils.dispatcher import Dispatcher, _read_yaml_file

from .client_cmd import validate_client_config
from .main import main
from .shared import apply_config


def get_statestore_config_from_file(init):
    """:param init:
    :return:
    """
    with open(init, "r") as file:
        try:
            settings = dict(yaml.safe_load(file))
            return settings
        except yaml.YAMLError as e:
            raise (e)


def check_helper_config_file(config):
    control = config["control"]
    try:
        helper = control["helper"]
    except KeyError:
        logger.error("--local-package was used, but no helper was found in --init settings file.")
        exit(-1)
    return helper


@main.group("run")
@click.pass_context
def run_cmd(ctx):
    """:param ctx:
    """
    pass


@run_cmd.command("build")
@click.option("-p", "--path", required=True, help="Path to package directory containing fedn.yaml")
@click.pass_context
def build_cmd(ctx, path):
    """Execute 'build' entrypoint in fedn.yaml.

    :param ctx:
    :param path: Path to folder containing fedn.yaml
    :type path: str
    """
    path = os.path.abspath(path)
    yaml_file = os.path.join(path, "fedn.yaml")
    if not os.path.exists(yaml_file):
        logger.error(f"Could not find fedn.yaml in {path}")
        exit(-1)

    config = _read_yaml_file(yaml_file)
    # Check that build is defined in fedn.yaml under entry_points
    if "build" not in config["entry_points"]:
        logger.error("No build command defined in fedn.yaml")
        exit(-1)

    dispatcher = Dispatcher(config, path)
    _ = dispatcher._get_or_create_python_env()
    dispatcher.run_cmd("build")

    # delete the virtualenv
    if dispatcher.python_env_path:
        logger.info(f"Removing virtualenv {dispatcher.python_env_path}")
        shutil.rmtree(dispatcher.python_env_path)


@run_cmd.command("client")
@click.option("-d", "--discoverhost", required=False, help="Hostname for discovery services(reducer).")
@click.option("-p", "--discoverport", required=False, help="Port for discovery services (reducer).")
@click.option("--token", required=False, help="Set token provided by reducer if enabled")
@click.option("-n", "--name", required=False, default="client" + str(uuid.uuid4())[:8])
@click.option("-i", "--client_id", required=False)
@click.option("--local-package", is_flag=True, help="Enable local compute package")
@click.option("--force-ssl", is_flag=True, help="Force SSL/TLS for REST service")
@click.option("-u", "--dry-run", required=False, default=False)
@click.option("-s", "--secure", required=False, default=False)
@click.option("-pc", "--preshared-cert", required=False, default=False)
@click.option("-v", "--verify", is_flag=True, help="Verify SSL/TLS for REST service")
@click.option("-c", "--preferred-combiner", required=False, default=False)
@click.option("-va", "--validator", required=False, default=True)
@click.option("-tr", "--trainer", required=False, default=True)
@click.option("-in", "--init", required=False, default=None, help="Set to a filename to (re)init client from file state.")
@click.option("-l", "--logfile", required=False, default=None, help="Set logfile for client log to file.")
@click.option("--heartbeat-interval", required=False, default=2)
@click.option("--reconnect-after-missed-heartbeat", required=False, default=30)
@click.option("--verbosity", required=False, default="INFO", type=click.Choice(["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"], case_sensitive=False))
@click.pass_context
def client_cmd(
    ctx,
    discoverhost,
    discoverport,
    token,
    name,
    client_id,
    local_package,
    force_ssl,
    dry_run,
    secure,
    preshared_cert,
    verify,
    preferred_combiner,
    validator,
    trainer,
    init,
    logfile,
    heartbeat_interval,
    reconnect_after_missed_heartbeat,
    verbosity,
):
    """:param ctx:
    :param discoverhost:
    :param discoverport:
    :param token:
    :param name:
    :param client_id:
    :param remote:
    :param dry_run:
    :param secure:
    :param preshared_cert:
    :param verify_cert:
    :param preferred_combiner:
    :param init:
    :param logfile:
    :param hearbeat_interval
    :param reconnect_after_missed_heartbeat
    :param verbosity
    :return:
    """
    remote = False if local_package else True
    config = {
        "discover_host": discoverhost,
        "discover_port": discoverport,
        "token": token,
        "name": name,
        "client_id": client_id,
        "remote_compute_context": remote,
        "force_ssl": force_ssl,
        "dry_run": dry_run,
        "secure": secure,
        "preshared_cert": preshared_cert,
        "verify": verify,
        "preferred_combiner": preferred_combiner,
        "validator": validator,
        "trainer": trainer,
        "logfile": logfile,
        "heartbeat_interval": heartbeat_interval,
        "reconnect_after_missed_heartbeat": reconnect_after_missed_heartbeat,
        "verbosity": verbosity,
    }

    click.echo(
        click.style("\n*** fedn run client is deprecated and will be removed. Please use fedn client start instead. ***\n", blink=True, bold=True, fg="red")
    )

    if init:
        apply_config(init, config)
        click.echo(f"\nClient configuration loaded from file: {init}")
        click.echo("Values set in file override defaults and command line arguments...\n")

    try:
        validate_client_config(config)
    except InvalidClientConfig as e:
        click.echo(f"Error: {e}")
        return

    client = Client(config)
    client.run()


@run_cmd.command("combiner")
@click.option("-d", "--discoverhost", required=False, help="Hostname for discovery services (reducer).")
@click.option("-p", "--discoverport", required=False, help="Port for discovery services (reducer).")
@click.option("-t", "--token", required=False, help="Set token provided by reducer if enabled")
@click.option("-n", "--name", required=False, default="combiner" + str(uuid.uuid4())[:8], help="Set name for combiner.")
@click.option("-h", "--host", required=False, default="combiner", help="Set hostname.")
@click.option("-i", "--port", required=False, default=12080, help="Set port.")
@click.option("-f", "--fqdn", required=False, default=None, help="Set fully qualified domain name")
@click.option("-s", "--secure", is_flag=True, help="Enable SSL/TLS encrypted gRPC channels.")
@click.option("-v", "--verify", is_flag=True, help="Verify SSL/TLS for REST discovery service (reducer)")
@click.option("-c", "--max_clients", required=False, default=30, help="The maximal number of client connections allowed.")
@click.option("-in", "--init", required=False, default=None, help="Path to configuration file to (re)init combiner.")
@click.pass_context
def combiner_cmd(ctx, discoverhost, discoverport, token, name, host, port, fqdn, secure, verify, max_clients, init):
    """:param ctx:
    :param discoverhost:
    :param discoverport:
    :param token:
    :param name:
    :param hostname:
    :param port:
    :param secure:
    :param max_clients:
    :param init:
    """
    config = {
        "discover_host": discoverhost,
        "discover_port": discoverport,
        "token": token,
        "host": host,
        "port": port,
        "fqdn": fqdn,
        "name": name,
        "secure": secure,
        "verify": verify,
        "max_clients": max_clients,
    }

    click.echo(
        click.style("\n*** fedn run combiner is deprecated and will be removed. Please use fedn combiner start instead. ***\n", blink=True, bold=True, fg="red")
    )

    if init:
        apply_config(init, config)
        click.echo(f"\nCombiner configuration loaded from file: {init}")
        click.echo("Values set in file override defaults and command line arguments...\n")

    combiner = Combiner(config)
    combiner.run()



import os

import requests

__all__ = ["APIClient"]


class APIClient:
    """An API client for interacting with the statestore and controller.

    :param host: The host of the api server.
    :type host: str
    :param port: The port of the api server.
    :type port: int
    :param secure: Whether to use https.
    :type secure: bool
    :param verify: Whether to verify the server certificate.
    :type verify: bool
    """

    def __init__(self, host, port=None, secure=False, verify=False, token=None, auth_scheme=None):
        self.host = host
        self.port = port
        self.secure = secure
        self.verify = verify
        self.headers = {}
        # Auth scheme passed as argument overrides environment variable.
        # "Token" is the default auth scheme.
        if not auth_scheme:
            auth_scheme = os.environ.get("FEDN_AUTH_SCHEME", "Bearer")
        # Override potential env variable if token is passed as argument.
        if not token:
            token = os.environ.get("FEDN_AUTH_TOKEN", False)

        if token:
            self.headers = {"Authorization": f"{auth_scheme} {token}"}

    def _get_url(self, endpoint):
        if self.secure:
            protocol = "https"
        else:
            protocol = "http"
        if self.port:
            return f"{protocol}://{self.host}:{self.port}/{endpoint}"
        return f"{protocol}://{self.host}/{endpoint}"

    def _get_url_api_v1(self, endpoint):
        return self._get_url(f"api/v1/{endpoint}")

    # --- Clients --- #

    def get_client(self, id: str):
        """Get a client from the statestore.

        :param id: The client id to get.
        :type id: str
        :return: Client.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1(f"clients/{id}"), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_clients(self, n_max: int = None):
        """Get clients from the statestore.

        :param n_max: The maximum number of clients to get (If none all will be fetched).
        :type n_max: int
        return: Clients.
        rtype: dict
        """
        _headers = self.headers.copy()

        if n_max:
            _headers["X-Limit"] = str(n_max)

        response = requests.get(self._get_url_api_v1("clients/"), verify=self.verify, headers=_headers)

        _json = response.json()

        return _json

    def get_clients_count(self):
        """Get the number of clients in the statestore.

        :return: The number of clients.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1("clients/count"), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_client_config(self, checksum=True):
        """Get client config from controller. Optionally include the checksum.
        The config is used for clients to connect to the controller and ask for combiner assignment.

        :param checksum: Whether to include the checksum of the package.
        :type checksum: bool
        :return: The client configuration.
        :rtype: dict
        """
        _params = {"checksum": "true" if checksum else "false"}

        response = requests.get(self._get_url("get_client_config"), params=_params, verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_active_clients(self, combiner_id: str = None, n_max: int = None):
        """Get active clients from the statestore.

        :param combiner_id: The combiner id to get active clients for.
        :type combiner_id: str
        :param n_max: The maximum number of clients to get (If none all will be fetched).
        :type n_max: int
        :return: Active clients.
        :rtype: dict
        """
        _params = {"status": "online"}

        if combiner_id:
            _params["combiner"] = combiner_id

        _headers = self.headers.copy()

        if n_max:
            _headers["X-Limit"] = str(n_max)

        response = requests.get(self._get_url_api_v1("clients/"), params=_params, verify=self.verify, headers=_headers)

        _json = response.json()

        return _json

    # --- Combiners --- #

    def get_combiner(self, id: str):
        """Get a combiner from the statestore.

        :param id: The combiner id to get.
        :type id: str
        :return: Combiner.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1(f"combiners/{id}"), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_combiners(self, n_max: int = None):
        """Get combiners in the network.

        :param n_max: The maximum number of combiners to get (If none all will be fetched).
        :type n_max: int
        :return: Combiners.
        :rtype: dict
        """
        _headers = self.headers.copy()

        if n_max:
            _headers["X-Limit"] = str(n_max)

        response = requests.get(self._get_url_api_v1("combiners/"), verify=self.verify, headers=_headers)

        _json = response.json()

        return _json

    def get_combiners_count(self):
        """Get the number of combiners in the statestore.

        :return: The number of combiners.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1("combiners/count"), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    # --- Controllers --- #

    def get_controller_status(self):
        """Get the status of the controller.

        :return: The status of the controller.
        :rtype: dict
        """
        response = requests.get(self._get_url("get_controller_status"), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    # --- Models --- #

    def get_model(self, id: str):
        """Get a model from the statestore.

        :param id: The id (or model property) of the model to get.
        :type id: str
        :return: Model.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1(f"models/{id}"), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_models(self, session_id: str = None, n_max: int = None):
        """Get models from the statestore.

        :param session_id: The session id to get models for. (optional)
        :type session_id: str
        :param n_max: The maximum number of models to get (If none all will be fetched).
        :type n_max: int
        :return: Models.
        :rtype: dict
        """
        _params = {}

        if session_id:
            _params["session_id"] = session_id

        _headers = self.headers.copy()

        if n_max:
            _headers["X-Limit"] = str(n_max)

        response = requests.get(self._get_url_api_v1("models/"), params=_params, verify=self.verify, headers=_headers)

        _json = response.json()

        return _json

    def get_models_count(self):
        """Get the number of models in the statestore.

        :return: The number of models.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1("models/count"), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_active_model(self):
        """Get the latest model from the statestore.

        :return: The latest model.
        :rtype: dict
        """
        _headers = self.headers.copy()
        _headers["X-Limit"] = "1"

        response = requests.get(self._get_url_api_v1("models/"), verify=self.verify, headers=_headers)
        _json = response.json()

        if "result" in _json and len(_json["result"]) > 0:
            return _json["result"][0]

        return _json

    def get_model_trail(self, id: str = None, include_self: bool = True, reverse: bool = True, n_max: int = None):
        """Get the model trail.

        :param id: The id (or model property) of the model to start the trail from. (optional)
        :type id: str
        :param n_max: The maximum number of models to get (If none all will be fetched).
        :type n_max: int
        :return: Models.
        :rtype: dict
        """
        if not id:
            model = self.get_active_model()
            if "id" in model:
                id = model["id"]
            else:
                return model

        _headers = self.headers.copy()

        _count: int = n_max if n_max else self.get_models_count()
        _headers["X-Limit"] = str(_count)
        _headers["X-Reverse"] = "true" if reverse else "false"

        _include_self_str: str = "true" if include_self else "false"

        response = requests.get(self._get_url_api_v1(f"models/{id}/ancestors?include_self={_include_self_str}"), verify=self.verify, headers=_headers)
        _json = response.json()

        return _json

    def download_model(self, id: str, path: str):
        """Download the model with id id.

        :param id: The id (or model property) of the model to download.
        :type id: str
        :param path: The path to download the model to.
        :type path: str
        :return: Message with success or failure.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1(f"models/{id}/download"), verify=self.verify, headers=self.headers)

        if response.status_code == 200:
            with open(path, "wb") as file:
                file.write(response.content)
            return {"success": True, "message": "Model downloaded successfully."}
        else:
            return {"success": False, "message": "Failed to download model."}

    def set_active_model(self, path):
        """Set the initial model in the statestore and upload to model repository.

        :param path: The file path of the initial model to set.
        :type path: str
        :return: A dict with success or failure message.
        :rtype: dict
        """
        with open(path, "rb") as file:
            response = requests.post(self._get_url("set_initial_model"), files={"file": file}, verify=self.verify, headers=self.headers)
        return response.json()

    # --- Packages --- #

    def get_package(self, id: str):
        """Get a compute package from the statestore.

        :param id: The id of the compute package to get.
        :type id: str
        :return: Package.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1(f"packages/{id}"), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_packages(self, n_max: int = None):
        """Get compute packages from the statestore.

        :param n_max: The maximum number of packages to get (If none all will be fetched).
        :type n_max: int
        :return: Packages.
        :rtype: dict
        """
        _headers = self.headers.copy()

        if n_max:
            _headers["X-Limit"] = str(n_max)

        response = requests.get(self._get_url_api_v1("packages/"), verify=self.verify, headers=_headers)

        _json = response.json()

        return _json

    def get_packages_count(self):
        """Get the number of compute packages in the statestore.

        :return: The number of packages.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1("packages/count"), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_active_package(self):
        """Get the (active) compute package from the statestore.

        :return: Package.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1("packages/active"), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_package_checksum(self):
        """Get the checksum of the compute package.

        :return: The checksum.
        :rtype: dict
        """
        response = requests.get(self._get_url("get_package_checksum"), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def download_package(self, path: str):
        """Download the compute package.

        :param path: The path to download the compute package to.
        :type path: str
        :return: Message with success or failure.
        :rtype: dict
        """
        response = requests.get(self._get_url("download_package"), verify=self.verify, headers=self.headers)
        if response.status_code == 200:
            with open(path, "wb") as file:
                file.write(response.content)
            return {"success": True, "message": "Package downloaded successfully."}
        else:
            return {"success": False, "message": "Failed to download package."}

    def set_active_package(self, path: str, helper: str, name: str = None, description: str = None):
        """Set the compute package in the statestore.

        :param path: The file path of the compute package to set.
        :type path: str
        :param helper: The helper type to use.
        :type helper: str
        :return: A dict with success or failure message.
        :rtype: dict
        """
        with open(path, "rb") as file:
            response = requests.post(
                self._get_url("set_package"),
                files={"file": file},
                data={"helper": helper, "name": name, "description": description},
                verify=self.verify,
                headers=self.headers,
            )

        _json = response.json()

        return _json

    # --- Rounds --- #

    def get_round(self, id: str):
        """Get a round from the statestore.

        :param round_id: The round id to get.
        :type round_id: str
        :return: Round (config and metrics).
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1(f"rounds/{id}"), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_rounds(self, n_max: int = None):
        """Get all rounds from the statestore.

        :param n_max: The maximum number of rounds to get (If none all will be fetched).
        :type n_max: int
        :return: Rounds.
        :rtype: dict
        """
        _headers = self.headers.copy()

        if n_max:
            _headers["X-Limit"] = str(n_max)

        response = requests.get(self._get_url_api_v1("rounds/"), verify=self.verify, headers=_headers)

        _json = response.json()

        return _json

    def get_rounds_count(self):
        """Get the number of rounds in the statestore.

        :return: The number of rounds.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1("rounds/count"), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    # --- Sessions --- #

    def get_session(self, id: str):
        """Get a session from the statestore.

        :param id: The session id to get.
        :type id: str
        :return: Session.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1(f"sessions/{id}"), self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_sessions(self, n_max: int = None):
        """Get sessions from the statestore.

        :param n_max: The maximum number of sessions to get (If none all will be fetched).
        :type n_max: int
        :return: Sessions.
        :rtype: dict
        """
        _headers = self.headers.copy()

        if n_max:
            _headers["X-Limit"] = str(n_max)

        response = requests.get(self._get_url_api_v1("sessions/"), verify=self.verify, headers=_headers)

        _json = response.json()

        return _json

    def get_sessions_count(self):
        """Get the number of sessions in the statestore.

        :return: The number of sessions.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1("sessions/count"), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_session_status(self, id: str):
        """Get the status of a session.

        :param id: The id of the session to get.
        :type id: str
        :return: The status of the session.
        :rtype: str
        """
        session = self.get_session(id)

        if session and "status" in session:
            return session["status"]

        return "Could not retrieve session status."

    def session_is_finished(self, id: str):
        """Check if a session with id has finished.

        :param id: The id of the session to get.
        :type id: str
        :return: True if session is finished, otherwise false.
        :rtype: bool
        """
        status = self.get_session_status(id)
        return status and status.lower() == "finished"

    def start_session(
        self,
        id: str = None,
        aggregator: str = "fedavg",
        aggregator_kwargs: dict = None,
        model_id: str = None,
        round_timeout: int = 180,
        rounds: int = 5,
        round_buffer_size: int = -1,
        delete_models: bool = True,
        validate: bool = True,
        helper: str = "",
        min_clients: int = 1,
        requested_clients: int = 8,
    ):
        """Start a new session.

        :param id: The session id to start.
        :type id: str
        :param aggregator: The aggregator plugin to use.
        :type aggregator: str
        :param model_id: The id of the initial model.
        :type model_id: str
        :param round_timeout: The round timeout to use in seconds.
        :type round_timeout: int
        :param rounds: The number of rounds to perform.
        :type rounds: int
        :param round_buffer_size: The round buffer size to use.
        :type round_buffer_size: int
        :param delete_models: Whether to delete models after each round at combiner (save storage).
        :type delete_models: bool
        :param validate: Whether to validate the model after each round.
        :type validate: bool
        :param helper: The helper type to use.
        :type helper: str
        :param min_clients: The minimum number of clients required.
        :type min_clients: int
        :param requested_clients: The requested number of clients.
        :type requested_clients: int
        :return: A dict with success or failure message and session config.
        :rtype: dict
        """
        response = requests.post(
            self._get_url("start_session"),
            json={
                "session_id": id,
                "aggregator": aggregator,
                "aggregator_kwargs": aggregator_kwargs,
                "model_id": model_id,
                "round_timeout": round_timeout,
                "rounds": rounds,
                "round_buffer_size": round_buffer_size,
                "delete_models": delete_models,
                "validate": validate,
                "helper": helper,
                "min_clients": min_clients,
                "requested_clients": requested_clients,
            },
            verify=self.verify,
            headers=self.headers,
        )

        _json = response.json()

        return _json

    # --- Statuses --- #

    def get_status(self, id: str):
        """Get a status object (event) from the statestore.

        :param id: The id of the status to get.
        :type id: str
        :return: Status.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1(f"statuses/{id}"), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_statuses(self, session_id: str = None, event_type: str = None, sender_name: str = None, sender_role: str = None, n_max: int = None):
        """Get statuses from the statestore. Filter by input parameters

        :param session_id: The session id to get statuses for.
        :type session_id: str
        :param event_type: The event type to get.
        :type event_type: str
        :param sender_name: The sender name to get.
        :type sender_name: str
        :param sender_role: The sender role to get.
        :type sender_role: str
        :param n_max: The maximum number of statuses to get (If none all will be fetched).
        :type n_max: int
        :return: Statuses
        """
        _params = {}

        if session_id:
            _params["session_id"] = session_id

        if event_type:
            _params["type"] = event_type

        if sender_name:
            _params["sender.name"] = sender_name

        if sender_role:
            _params["sender.role"] = sender_role

        _headers = self.headers.copy()

        if n_max:
            _headers["X-Limit"] = str(n_max)

        response = requests.get(self._get_url_api_v1("statuses/"), params=_params, verify=self.verify, headers=_headers)

        _json = response.json()

        return _json

    def get_statuses_count(self):
        """Get the number of statuses in the statestore.

        :return: The number of statuses.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1("statuses/count"), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    # --- Validations --- #

    def get_validation(self, id: str):
        """Get a validation from the statestore.

        :param id: The id of the validation to get.
        :type id: str
        :return: Validation.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1(f"validations/{id}"), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json

    def get_validations(
        self,
        session_id: str = None,
        model_id: str = None,
        correlation_id: str = None,
        sender_name: str = None,
        sender_role: str = None,
        receiver_name: str = None,
        receiver_role: str = None,
        n_max: int = None,
    ):
        """Get validations from the statestore. Filter by input parameters.

        :param session_id: The session id to get validations for.
        :type session_id: str
        :param model_id: The model id to get validations for.
        :type model_id: str
        :param correlation_id: The correlation id to get validations for.
        :type correlation_id: str
        :param sender_name: The sender name to get validations for.
        :type sender_name: str
        :param sender_role: The sender role to get validations for.
        :type sender_role: str
        :param receiver_name: The receiver name to get validations for.
        :type receiver_name: str
        :param receiver_role: The receiver role to get validations for.
        :type receiver_role: str
        :param n_max: The maximum number of validations to get (If none all will be fetched).
        :type n_max: int
        :return: Validations.
        :rtype: dict
        """
        _params = {}

        if session_id:
            _params["sessionId"] = session_id

        if model_id:
            _params["modelId"] = model_id

        if correlation_id:
            _params["correlationId"] = correlation_id

        if sender_name:
            _params["sender.name"] = sender_name

        if sender_role:
            _params["sender.role"] = sender_role

        if receiver_name:
            _params["receiver.name"] = receiver_name

        if receiver_role:
            _params["receiver.role"] = receiver_role

        _headers = self.headers.copy()

        if n_max:
            _headers["X-Limit"] = str(n_max)

        response = requests.get(self._get_url_api_v1("validations/"), params=_params, verify=self.verify, headers=_headers)

        _json = response.json()

        return _json

    def get_validations_count(self):
        """Get the number of validations in the statestore.

        :return: The number of validations.
        :rtype: dict
        """
        response = requests.get(self._get_url_api_v1("validations/count"), verify=self.verify, headers=self.headers)

        _json = response.json()

        return _json



import importlib.metadata

import fedn


def get_version(pacakge):
    # Dynamically get the version of the package
    try:
        version = importlib.metadata.version("fedn")
    except importlib.metadata.PackageNotFoundError:
        version = "unknown"
    return version


def get_package_path():
    # Get the path of the package
    return fedn.__path__[0]



import io

import numpy as np
from flask import Blueprint, jsonify, request, send_file

from fedn.network.api.auth import jwt_auth_required
from fedn.network.api.shared import modelstorage_config
from fedn.network.api.v1.shared import api_version, get_limit, get_post_data_to_kwargs, get_reverse, get_typed_list_headers, mdb
from fedn.network.storage.s3.base import RepositoryBase
from fedn.network.storage.s3.miniorepository import MINIORepository
from fedn.network.storage.statestore.stores.model_store import ModelStore
from fedn.network.storage.statestore.stores.shared import EntityNotFound

bp = Blueprint("model", __name__, url_prefix=f"/api/{api_version}/models")

model_store = ModelStore(mdb, "control.model")
repository: RepositoryBase = None

if modelstorage_config["storage_type"] == "S3":
    repository = MINIORepository(modelstorage_config["storage_config"])


@bp.route("/", methods=["GET"])
@jwt_auth_required(role="admin")
def get_models():
    """Get models
    Retrieves a list of models based on the provided parameters.
    By specifying a parameter in the url, you can filter the models based on that parameter,
    and the response will contain only the models that match the filter.
    ---
    tags:
      - Models
    parameters:
      - in: query
        name: model
        schema:
        type: string
        description: A unique identifier for the model
      - in: query
        name: parent_model
        schema:
        type: string
        description: The unique identifier of the parent model
      - in: query
        name: session_id
        schema:
        type: string
        description: The unique identifier of the session
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of models to retrieve
      - name: X-Skip
        in: header
        required: false
        type: integer
        description: The number of models to skip
      - name: X-Sort-Key
        in: header
        required: false
        type: string
        description: The key to sort the models by
      - name: X-Sort-Order
        in: header
        required: false
        type: string
        description: The order to sort the models in ('asc' or 'desc')
    definitions:
        Model:
            type: object
            properties:
            id:
                type: string
                description: The id of the model
            model:
                type: string
                description: A unique identifier for the model
            parent_model:
                type: string
                description: The unique identifier of the parent model
            session_id:
                type: string
                description: The unique identifier of the session
            committed_at:
                type: string
                format: date-time
                description: The date and time the model was created
    responses:
      200:
        description: A list of models and the total count.
        schema:
            type: object
            properties:
                count:
                    type: integer
                result:
                    type: array
                    items:
                        $ref: '#/definitions/Model'
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                error:
                    type: string
    """
    try:
        limit, skip, sort_key, sort_order, _ = get_typed_list_headers(request.headers)
        kwargs = request.args.to_dict()

        models = model_store.list(limit, skip, sort_key, sort_order, use_typing=False, **kwargs)

        result = models["result"]

        response = {"count": models["count"], "result": result}

        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/list", methods=["POST"])
@jwt_auth_required(role="admin")
def list_models():
    """List models
    Retrieves a list of models based on the provided parameters.
    Works much like the GET /models endpoint, but allows for more complex queries.
    By specifying a parameter in the body, you can filter the models based on that parameter,
    and the response will contain only the models that match the filter. If the parameter value contains a comma,
    the filter will be an "in" query, meaning that the models will be returned if the specified field contains any of the values in the parameter.
    ---
    tags:
      - Models
    parameters:
      - name: model
        in: body
        required: false
        type: object
        description: Object containing the model filter
        schema:
          type: object
          properties:
            model:
              type: string
              description: A unique identifier for the model
            parent_model:
              type: string
              description: The unique identifier of the parent model
            session_id:
              type: string
              description: The unique identifier of the session
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of models to retrieve
      - name: X-Skip
        in: header
        required: false
        type: integer
        description: The number of models to skip
      - name: X-Sort-Key
        in: header
        required: false
        type: string
        description: The key to sort the models by
      - name: X-Sort-Order
        in: header
        required: false
        type: string
        description: The order to sort the models in ('asc' or 'desc')
    responses:
      200:
        description: A list of models and the total count.
        schema:
            type: object
            properties:
                count:
                    type: integer
                result:
                    type: array
                    items:
                        $ref: '#/definitions/Model'
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                error:
                    type: string
    """
    try:
        limit, skip, sort_key, sort_order, _ = get_typed_list_headers(request.headers)
        kwargs = get_post_data_to_kwargs(request)

        models = model_store.list(limit, skip, sort_key, sort_order, use_typing=False, **kwargs)

        result = models["result"]

        response = {"count": models["count"], "result": result}

        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/count", methods=["GET"])
@jwt_auth_required(role="admin")
def get_models_count():
    """Models count
    Retrieves the count of models based on the provided parameters.
    By specifying a parameter in the url, you can filter the models based on that parameter,
    and the response will contain only the count of models that match the filter.
    ---
    tags:
        - Models
    parameters:
      - in: query
        name: model
        schema:
        type: string
        description: A unique identifier for the model
      - in: query
        name: parent_model
        schema:
        type: string
        description: The unique identifier of the parent model
      - in: query
        name: session_id
        schema:
        type: string
        description: The unique identifier of the session
    responses:
        200:
            description: The count of models
            schema:
                type: integer
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        kwargs = request.args.to_dict()
        count = model_store.count(**kwargs)
        response = count
        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/count", methods=["POST"])
@jwt_auth_required(role="admin")
def models_count():
    """Models count
    Retrieves the count of models based on the provided parameters.
    Much like the GET /models/count endpoint, but allows for more complex queries.
    By specifying a parameter in the body, you can filter the models based on that parameter,
    and the response will contain only the count of models that match the filter. If the parameter value contains a comma,
    the filter will be an "in" query, meaning that the models will be returned if the specified field contains any of the values in the parameter.
    ---
    tags:
        - Models
    parameters:
      - name: model
        in: body
        required: false
        type: object
        description: Object containing the model filter
        schema:
          type: object
          properties:
            model:
              type: string
              description: A unique identifier for the model
            parent_model:
              type: string
              description: The unique identifier of the parent model
            session_id:
              type: string
              description: The unique identifier of the session
    responses:
        200:
            description: The count of models
            schema:
                type: integer
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        kwargs = get_post_data_to_kwargs(request)
        count = model_store.count(**kwargs)
        response = count
        return jsonify(response), 200
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/<string:id>", methods=["GET"])
@jwt_auth_required(role="admin")
def get_model(id: str):
    """Get model
    Retrieves a model based on the provided id.
    ---
    tags:
        - Models
    parameters:
      - name: id
        in: path
        required: true
        type: string
        description: The id or model property of the model
    responses:
        200:
            description: The model
            schema:
                $ref: '#/definitions/Model'
        404:
            description: The model was not found
            schema:
                type: object
                properties:
                    message:
                        type: string
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        model = model_store.get(id, use_typing=False)

        response = model

        return jsonify(response), 200
    except EntityNotFound:
        return jsonify({"message": f"Entity with id: {id} not found"}), 404
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/<string:id>", methods=["PATCH"])
@jwt_auth_required(role="admin")
def patch_model(id: str):
    """Patch model
    Updates a model based on the provided id. Only the fields that are present in the request will be updated.
    ---
    tags:
        - Models
    parameters:
        - name: id
            in: path
            required: true
            type: string
            description: The id or model property of the model
        - name: model
            in: body
            required: true
            type: object
            description: The model data to update
    responses:
        200:
            description: The updated model
            schema:
                $ref: '#/definitions/Model'
        404:
            description: The model was not found
            schema:
                type: object
                properties:
                    message:
                        type: string
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        model = model_store.get(id, use_typing=False)

        data = request.get_json()
        _id = model["id"]

        # Update the model with the new data
        # Only update the fields that are present in the request
        for key, value in data.items():
            if key in ["_id", "model"]:
                continue
            model[key] = value

        success, message = model_store.update(_id, model)

        if success:
            response = model
            return jsonify(response), 200

        return jsonify({"message": f"Failed to update model: {message}"}), 500
    except EntityNotFound:
        return jsonify({"message": f"Entity with id: {id} not found"}), 404
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/<string:id>", methods=["PUT"])
@jwt_auth_required(role="admin")
def put_model(id: str):
    """Put model
    Updates a model based on the provided id. All fields will be updated with the new data.
    ---
    tags:
        - Models
    parameters:
        - name: id
            in: path
            required: true
            type: string
            description: The id or model property of the model
        - name: model
            in: body
            required: true
            type: object
            description: The model data to update
    responses:
        200:
            description: The updated model
            schema:
                $ref: '#/definitions/Model'
        404:
            description: The model was not found
            schema:
                type: object
                properties:
                    message:
                        type: string
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        model = model_store.get(id, use_typing=False)
        data = request.get_json()
        _id = model["id"]

        success, message = model_store.update(_id, data)

        if success:
            response = model
            return jsonify(response), 200

        return jsonify({"message": f"Failed to update model: {message}"}), 500
    except EntityNotFound:
        return jsonify({"message": f"Entity with id: {id} not found"}), 404
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/<string:id>/descendants", methods=["GET"])
@jwt_auth_required(role="admin")
def get_descendants(id: str):
    """Get model descendants
    Retrieves a list of model descendants of the provided model id/model property.
    ---
    tags:
        - Models
    parameters:
      - name: id
        in: path
        required: true
        type: string
        description: The id or model property of the model
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of models to retrieve (defaults to 10)
        default: 10
    responses:
        200:
            description: The model
            schema:
                $ref: '#/definitions/Model'
        404:
            description: The inital model was not found
            schema:
                type: object
                properties:
                    message:
                        type: string
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        limit = get_limit(request.headers)

        descendants = model_store.list_descendants(id, limit or 10, use_typing=False)

        response = descendants

        return jsonify(response), 200
    except EntityNotFound:
        return jsonify({"message": f"Entity with id: {id} not found"}), 404
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/<string:id>/ancestors", methods=["GET"])
@jwt_auth_required(role="admin")
def get_ancestors(id: str):
    """Get model ancestors
    Retrieves a list of model ancestors of the provided model id/model property.
    ---
    tags:
        - Models
    parameters:
      - name: id
        in: path
        required: true
        type: string
        description: The id or model property of the model
      - name: include_self
        in: query
        required: false
        type: boolean
        description: Whether to include the initial model in the response
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of models to retrieve (defaults to 10)
        default: 10
      - name: X-Reverse
        in: header
        required: false
        type: boolean
        description: Whether to reverse the order of the ancestors
        default: false
    responses:
        200:
            description: The model
            schema:
                $ref: '#/definitions/Model'
        404:
            description: The inital model was not found
            schema:
                type: object
                properties:
                    message:
                        type: string
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        limit = get_limit(request.headers)
        reverse = get_reverse(request.headers)
        include_self_param: str = request.args.get("include_self")

        include_self: bool = include_self_param and include_self_param.lower() == "true"

        ancestors = model_store.list_ancestors(id, limit or 10, include_self=include_self, reverse=reverse, use_typing=False)

        response = ancestors

        return jsonify(response), 200
    except EntityNotFound:
        return jsonify({"message": f"Entity with id: {id} not found"}), 404
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/<string:id>/download", methods=["GET"])
@jwt_auth_required(role="admin")
def download(id: str):
    """Download
    Downloads the model file of the provided id.
    ---
    tags:
        - Models
    parameters:
      - name: id
        in: path
        required: true
        type: string
        description: The id or model property of the model
    responses:
        200:
            type: file
            description: The model file
        404:
            description: The model was not found
            schema:
                type: object
                properties:
                    message:
                        type: string
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        if repository is not None:
            model = model_store.get(id, use_typing=False)
            model_id = model["model"]

            file = repository.get_artifact_stream(model_id, modelstorage_config["storage_config"]["storage_bucket"])

            return send_file(file, as_attachment=True, download_name=model_id)
        else:
            return jsonify({"message": "No model storage configured"}), 500
    except EntityNotFound:
        return jsonify({"message": f"Entity with id: {id} not found"}), 404
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/<string:id>/parameters", methods=["GET"])
@jwt_auth_required(role="admin")
def get_parameters(id: str):
    """Download
    Downloads parameters of the model of the provided id.
    Please not that this endpoint is only available for models that have been stored as numpy arrays.
    ---
    tags:
        - Models
    parameters:
      - name: id
        in: path
        required: true
        type: string
        description: The id or model property of the model
    responses:
        200:
            description: The model parameters
            schema:
                type: object
                properties:
                    parameters:
                        type: object (array of arrays)
        404:
            description: The model was not found
            schema:
                type: object
                properties:
                    message:
                        type: string
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        if repository is not None:
            model = model_store.get(id, use_typing=False)
            model_id = model["model"]

            file = repository.get_artifact_stream(model_id, modelstorage_config["storage_config"]["storage_bucket"])

            file_bytes = io.BytesIO()
            for chunk in file.stream(32 * 1024):
                file_bytes.write(chunk)
            file_bytes.seek(0)  # Reset the pointer to the beginning of the byte array

            a = np.load(file_bytes)

            weights = []
            for i in range(len(a.files)):
                weights.append(a[str(i)].tolist())

            return jsonify(array=weights), 200
        else:
            return jsonify({"message": "No model storage configured"}), 500
    except EntityNotFound:
        return jsonify({"message": f"Entity with id: {id} not found"}), 404
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/active", methods=["GET"])
@jwt_auth_required(role="admin")
def get_active_model():
    """Get active model
    Retrieves the active model (id).
    ---
    tags:
        - Models
    responses:
        200:
            description: The active model id
            schema:
                type: string
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        active_model = model_store.get_active()

        response = active_model

        return jsonify(response), 200
    except EntityNotFound:
        return jsonify({"message": "No active model found"}), 404
    except Exception:
        return jsonify({"message": "An unexpected error occurred"}), 500
